@@ -16,7 +16,7 @@
 
 """Helper functions."""
 
-from __future__ import absolute_import, division, unicode_literals
+from __future__ import absolute_import, annotations, division, unicode_literals
 
 import base64
 import errno
@@ -26,14 +26,18 @@ import os
 import platform
 import re
 import urllib.parse
+from datetime import datetime
 from queue import Queue
 from threading import BoundedSemaphore, Thread
+from typing import BinaryIO, Dict, List, Mapping, Tuple, Union
+
+from typing_extensions import Protocol
+from urllib3._collections import HTTPHeaderDict
 
 from . import __title__, __version__
 from .sse import Sse, SseCustomerKey
 from .time import to_iso8601utc
 
-# Constants
 _DEFAULT_USER_AGENT = (
     f"MinIO ({platform.system()}; {platform.machine()}) "
     f"{__title__}/{__version__}"
@@ -75,8 +79,15 @@ _AWS_S3_PREFIX_REGEX = re.compile(_AWS_S3_PREFIX, re.IGNORECASE)
 _REGION_REGEX = re.compile(r'^((?!_)(?!-)[a-z_\d-]{1,63}(?<!-)(?<!_))$',
                            re.IGNORECASE)
 
+DictType = Dict[str, Union[str, List[str], Tuple[str]]]
 
-def quote(resource, safe='/', encoding=None, errors=None):
+
+def quote(
+        resource: str,
+        safe: str = "/",
+        encoding: str | None = None,
+        errors: str | None = None,
+) -> str:
     """
     Wrapper to urllib.parse.quote() replacing back to '~' for older python
     versions.
@@ -89,45 +100,44 @@ def quote(resource, safe='/', encoding=None, errors=None):
     ).replace("%7E", "~")
 
 
-def queryencode(query, safe='', encoding=None, errors=None):
+def queryencode(
+        query: str,
+        safe: str = "",
+        encoding: str | None = None,
+        errors: str | None = None,
+) -> str:
     """Encode query parameter value."""
     return quote(query, safe, encoding, errors)
 
 
-def headers_to_strings(headers, titled_key=False):
+def headers_to_strings(
+        headers: Mapping[str, str | list[str] | tuple[str]],
+        titled_key: bool = False,
+) -> str:
     """Convert HTTP headers to multi-line string."""
-    def _get_key(key):
-        return key.title() if titled_key else key
-
-    def _get_value(value):
-        return re.sub(
-            r"Credential=([^/]+)",
-            "Credential=*REDACTED*",
-            re.sub(
-                r"Signature=([0-9a-f]+)",
-                "Signature=*REDACTED*",
-                value if isinstance(value, str) else str(value),
-            ),
-        ) if titled_key else value
-
-    return "\n".join(
-        [
-            f"{_get_key(key)}: {_get_value(value)}"
-            for key, value in headers.items()
-        ]
-    )
+    values = []
+    for key, value in headers.items():
+        key = key.title() if titled_key else key
+        for item in value if isinstance(value, (list, tuple)) else [value]:
+            item = re.sub(
+                r"Credential=([^/]+)",
+                "Credential=*REDACTED*",
+                re.sub(r"Signature=([0-9a-f]+)", "Signature=*REDACTED*", item),
+            ) if titled_key else item
+            values.append(f"{key}: {item}")
+    return "\n".join(values)
 
 
-def _validate_sizes(object_size, part_size):
+def _validate_sizes(object_size: int, part_size: int):
     """Validate object and part size."""
     if part_size > 0:
         if part_size < MIN_PART_SIZE:
             raise ValueError(
-                f"part size {part_size} is not supported; minimum allowed 5MiB",
+                f"part size {part_size} is not supported; minimum allowed 5MiB"
             )
         if part_size > MAX_PART_SIZE:
             raise ValueError(
-                f"part size {part_size} is not supported; maximum allowed 5GiB",
+                f"part size {part_size} is not supported; maximum allowed 5GiB"
             )
 
     if object_size >= 0:
@@ -142,7 +152,7 @@ def _validate_sizes(object_size, part_size):
         )
 
 
-def _get_part_info(object_size, part_size):
+def _get_part_info(object_size: int, part_size: int):
     """Compute part information for object and part size."""
     _validate_sizes(object_size, part_size)
 
@@ -159,7 +169,7 @@ def _get_part_info(object_size, part_size):
     return part_size, math.ceil(object_size / part_size) if part_size else 1
 
 
-def get_part_info(object_size, part_size):
+def get_part_info(object_size: int, part_size: int) -> tuple[int, int]:
     """Compute part information for object and part size."""
     part_size, part_count = _get_part_info(object_size, part_size)
     if part_count > MAX_MULTIPART_COUNT:
@@ -170,7 +180,22 @@ def get_part_info(object_size, part_size):
     return part_size, part_count
 
 
-def read_part_data(stream, size, part_data=b'', progress=None):
+class ProgressType(Protocol):
+    """typing stub for Put/Get object progress."""
+
+    def set_meta(self, object_name: str, total_length: int):
+        """Set process meta information."""
+
+    def update(self, length: int):
+        """Set current progress length."""
+
+
+def read_part_data(
+        stream: BinaryIO,
+        size: int,
+        part_data: bytes = b"",
+        progress: ProgressType | None = None,
+) -> bytes:
     """Read part data of given size from stream."""
     size -= len(part_data)
     while size:
@@ -186,7 +211,7 @@ def read_part_data(stream, size, part_data=b'', progress=None):
     return part_data
 
 
-def makedirs(path):
+def makedirs(path: str):
     """Wrapper of os.makedirs() ignores errno.EEXIST."""
     try:
         if path:
@@ -199,7 +224,11 @@ def makedirs(path):
             raise ValueError(f"path {path} is not a directory") from exc
 
 
-def check_bucket_name(bucket_name, strict=False, s3_check=False):
+def check_bucket_name(
+        bucket_name: str,
+        strict: bool = False,
+        s3_check: bool = False,
+):
     """Check whether bucket name is valid optional with strict check or not."""
 
     if strict:
@@ -229,7 +258,7 @@ def check_bucket_name(bucket_name, strict=False, s3_check=False):
                          "'--ol-s3'")
 
 
-def check_non_empty_string(string):
+def check_non_empty_string(string: str | bytes):
     """Check whether given string is not empty."""
     try:
         if not string.strip():
@@ -238,7 +267,7 @@ def check_non_empty_string(string):
         raise TypeError() from exc
 
 
-def is_valid_policy_type(policy):
+def is_valid_policy_type(policy: str | bytes):
     """
     Validate if policy is type str
 
@@ -254,43 +283,53 @@ def is_valid_policy_type(policy):
     return True
 
 
-def check_ssec(sse):
+def check_ssec(sse: SseCustomerKey | None):
     """Check sse is SseCustomerKey type or not."""
     if sse and not isinstance(sse, SseCustomerKey):
         raise ValueError("SseCustomerKey type is required")
 
 
-def check_sse(sse):
+def check_sse(sse: Sse | None):
     """Check sse is Sse type or not."""
     if sse and not isinstance(sse, Sse):
         raise ValueError("Sse type is required")
 
 
-def md5sum_hash(data):
+def md5sum_hash(data: str | bytes | None) -> str | None:
     """Compute MD5 of data and return hash as Base64 encoded value."""
     if data is None:
         return None
 
     # indicate md5 hashing algorithm is not used in a security context.
     # Refer https://bugs.python.org/issue9216 for more information.
-    hasher = hashlib.new("md5", usedforsecurity=False)
+    hasher = hashlib.new(  # type: ignore[call-arg]
+        "md5",
+        usedforsecurity=False,
+    )
     hasher.update(data.encode() if isinstance(data, str) else data)
     md5sum = base64.b64encode(hasher.digest())
     return md5sum.decode() if isinstance(md5sum, bytes) else md5sum
 
 
-def sha256_hash(data):
+def sha256_hash(data: str | bytes | None) -> str:
     """Compute SHA-256 of data and return hash as hex encoded value."""
     data = data or b""
     hasher = hashlib.sha256()
     hasher.update(data.encode() if isinstance(data, str) else data)
     sha256sum = hasher.hexdigest()
-    return sha256sum.decode() if isinstance(sha256sum, bytes) else sha256sum
+    if isinstance(sha256sum, bytes):
+        return sha256sum.decode()
+    return sha256sum
 
 
 def url_replace(
-        url, scheme=None, netloc=None, path=None, query=None, fragment=None
-):
+        url: urllib.parse.SplitResult,
+        scheme: str | None = None,
+        netloc: str | None = None,
+        path: str | None = None,
+        query: str | None = None,
+        fragment: str | None = None,
+) -> urllib.parse.SplitResult:
     """Return new URL with replaced properties in given URL."""
     return urllib.parse.SplitResult(
         scheme if scheme is not None else url.scheme,
@@ -301,14 +340,14 @@ def url_replace(
     )
 
 
-def _metadata_to_headers(metadata):
+def _metadata_to_headers(metadata: DictType) -> dict[str, list[str]]:
     """Convert user metadata to headers."""
-    def normalize_key(key):
+    def normalize_key(key: str) -> str:
         if not key.lower().startswith("x-amz-meta-"):
             key = "X-Amz-Meta-" + key
         return key
 
-    def to_string(value):
+    def to_string(value) -> str:
         value = str(value)
         try:
             value.encode("us-ascii")
@@ -319,7 +358,7 @@ def _metadata_to_headers(metadata):
             ) from exc
         return value
 
-    def normalize_value(values):
+    def normalize_value(values: str | list[str] | tuple[str]) -> list[str]:
         if not isinstance(values, (list, tuple)):
             values = [values]
         return [to_string(value) for value in values]
@@ -330,11 +369,11 @@ def _metadata_to_headers(metadata):
     }
 
 
-def normalize_headers(headers):
+def normalize_headers(headers: DictType | None) -> DictType:
     """Normalize headers by prefixing 'X-Amz-Meta-' for user metadata."""
     headers = {str(key): value for key, value in (headers or {}).items()}
 
-    def guess_user_metadata(key):
+    def guess_user_metadata(key: str) -> bool:
         key = key.lower()
         return not (
             key.startswith("x-amz-") or
@@ -359,7 +398,13 @@ def normalize_headers(headers):
     return headers
 
 
-def genheaders(headers, sse, tags, retention, legal_hold):
+def genheaders(
+        headers: DictType | None,
+        sse: Sse | None,
+        tags: dict[str, str] | None,
+        retention,
+        legal_hold: bool,
+) -> DictType:
     """Generate headers for given parameters."""
     headers = normalize_headers(headers)
     headers.update(sse.headers() if sse else {})
@@ -374,14 +419,18 @@ def genheaders(headers, sse, tags, retention, legal_hold):
     if retention and retention.mode:
         headers["x-amz-object-lock-mode"] = retention.mode
         headers["x-amz-object-lock-retain-until-date"] = (
-            to_iso8601utc(retention.retain_until_date)
+            to_iso8601utc(retention.retain_until_date) or ""
         )
     if legal_hold:
         headers["x-amz-object-lock-legal-hold"] = "ON"
     return headers
 
 
-def _get_aws_info(host, https, region):
+def _get_aws_info(
+        host: str,
+        https: bool,
+        region: str | None,
+) -> tuple[dict | None, str | None]:
     """Extract AWS domain information. """
 
     if not _HOSTNAME_REGEX.match(host):
@@ -394,10 +443,14 @@ def _get_aws_info(host, https, region):
     if not _AWS_ENDPOINT_REGEX.match(host):
         return (None, None)
 
+    if host.startswith("ec2-"):
+        return (None, None)
+
     if not _AWS_S3_ENDPOINT_REGEX.match(host):
         raise ValueError(f"invalid Amazon AWS host {host}")
 
-    end = _AWS_S3_PREFIX_REGEX.match(host).end()
+    matcher = _AWS_S3_PREFIX_REGEX.match(host)
+    end = matcher.end() if matcher else 0
     aws_s3_prefix = host[:end]
 
     if "s3-accesspoint" in aws_s3_prefix and not https:
@@ -407,7 +460,7 @@ def _get_aws_info(host, https, region):
     dualstack = tokens[0] == "dualstack"
     if dualstack:
         tokens = tokens[1:]
-    region_in_host = None
+    region_in_host = ""
     if tokens[0] not in ["vpce", "amazonaws"]:
         region_in_host = tokens[0]
         tokens = tokens[1:]
@@ -434,7 +487,7 @@ def _get_aws_info(host, https, region):
              "dualstack": dualstack}, None)
 
 
-def _parse_url(endpoint):
+def _parse_url(endpoint: str) -> urllib.parse.SplitResult:
     """Parse url string."""
 
     url = urllib.parse.urlsplit(endpoint)
@@ -478,17 +531,23 @@ def _parse_url(endpoint):
 
 class BaseURL:
     """Base URL of S3 endpoint."""
+    _aws_info: dict | None
+    _virtual_style_flag: bool
+    _url: urllib.parse.SplitResult
+    _region: str | None
+    _accelerate_host_flag: bool
 
-    def __init__(self, endpoint, region):
+    def __init__(self, endpoint: str, region: str | None):
         url = _parse_url(endpoint)
 
         if region and not _REGION_REGEX.match(region):
             raise ValueError(f"invalid region {region}")
 
+        hostname = url.hostname or ""
         self._aws_info, region_in_host = _get_aws_info(
-            url.hostname, url.scheme == "https", region)
+            hostname, url.scheme == "https", region)
         self._virtual_style_flag = (
-            self._aws_info or url.hostname.endswith("aliyuncs.com")
+            self._aws_info is not None or hostname.endswith("aliyuncs.com")
         )
         self._url = url
         self._region = region or region_in_host
@@ -500,32 +559,32 @@ class BaseURL:
             )
 
     @property
-    def region(self):
+    def region(self) -> str | None:
         """Get region."""
         return self._region
 
     @property
-    def is_https(self):
+    def is_https(self) -> bool:
         """Check if scheme is HTTPS."""
         return self._url.scheme == "https"
 
     @property
-    def host(self):
+    def host(self) -> str:
         """Get hostname."""
         return self._url.netloc
 
     @property
-    def is_aws_host(self):
+    def is_aws_host(self) -> bool:
         """Check if URL points to AWS host."""
         return self._aws_info is not None
 
     @property
-    def aws_s3_prefix(self):
+    def aws_s3_prefix(self) -> str | None:
         """Get AWS S3 domain prefix."""
         return self._aws_info["s3_prefix"] if self._aws_info else None
 
     @aws_s3_prefix.setter
-    def aws_s3_prefix(self, s3_prefix):
+    def aws_s3_prefix(self, s3_prefix: str):
         """Set AWS s3 domain prefix."""
         if not _AWS_S3_PREFIX_REGEX.match(s3_prefix):
             raise ValueError(f"invalid AWS S3 domain prefix {s3_prefix}")
@@ -533,40 +592,48 @@ class BaseURL:
             self._aws_info["s3_prefix"] = s3_prefix
 
     @property
-    def accelerate_host_flag(self):
+    def accelerate_host_flag(self) -> bool:
         """Get AWS accelerate host flag."""
         return self._accelerate_host_flag
 
     @accelerate_host_flag.setter
-    def accelerate_host_flag(self, flag):
+    def accelerate_host_flag(self, flag: bool):
         """Set AWS accelerate host flag."""
         self._accelerate_host_flag = flag
 
     @property
-    def dualstack_host_flag(self):
+    def dualstack_host_flag(self) -> bool:
         """Check if URL points to AWS dualstack host."""
         return self._aws_info["dualstack"] if self._aws_info else False
 
     @dualstack_host_flag.setter
-    def dualstack_host_flag(self, flag):
+    def dualstack_host_flag(self, flag: bool):
         """Set AWS dualstack host."""
         if self._aws_info:
             self._aws_info["dualstack"] = flag
 
     @property
-    def virtual_style_flag(self):
+    def virtual_style_flag(self) -> bool:
         """Check to use virtual style or not."""
         return self._virtual_style_flag
 
     @virtual_style_flag.setter
-    def virtual_style_flag(self, flag):
+    def virtual_style_flag(self, flag: bool):
         """Check to use virtual style or not."""
         self._virtual_style_flag = flag
 
-    def _build_aws_url(self, url, bucket_name, enforce_path_style, region):
+    @classmethod
+    def _build_aws_url(
+            cls,
+            aws_info: dict,
+            url: urllib.parse.SplitResult,
+            bucket_name: str | None,
+            enforce_path_style: bool,
+            region: str,
+    ) -> urllib.parse.SplitResult:
         """Build URL for given information."""
-        s3_prefix = self._aws_info["s3_prefix"]
-        domain_suffix = self._aws_info["domain_suffix"]
+        s3_prefix = aws_info["s3_prefix"]
+        domain_suffix = aws_info["domain_suffix"]
 
         host = f"{s3_prefix}{domain_suffix}"
         if host in ["s3-external-1.amazonaws.com",
@@ -576,7 +643,7 @@ class BaseURL:
 
         netloc = s3_prefix
         if "s3-accelerate" in s3_prefix:
-            if "." in bucket_name:
+            if "." in (bucket_name or ""):
                 raise ValueError(
                     f"bucket name '{bucket_name}' with '.' is not allowed "
                     f"for accelerate endpoint"
@@ -584,7 +651,7 @@ class BaseURL:
             if enforce_path_style:
                 netloc = netloc.replace("-accelerate", "", 1)
 
-        if self._aws_info["dualstack"]:
+        if aws_info["dualstack"]:
             netloc += "dualstack."
         if "s3-accelerate" not in s3_prefix:
             netloc += region + "."
@@ -592,7 +659,11 @@ class BaseURL:
 
         return url_replace(url, netloc=netloc)
 
-    def _build_list_buckets_url(self, url, region):
+    def _build_list_buckets_url(
+            self,
+            url: urllib.parse.SplitResult,
+            region: str | None,
+    ) -> urllib.parse.SplitResult:
         """Build URL for ListBuckets API."""
         if not self._aws_info:
             return url
@@ -613,9 +684,13 @@ class BaseURL:
         return url_replace(url, netloc=f"{s3_prefix}{region}.{domain_suffix}")
 
     def build(
-            self, method, region,
-            bucket_name=None, object_name=None, query_params=None,
-    ):
+            self,
+            method: str,
+            region: str,
+            bucket_name: str | None = None,
+            object_name: str | None = None,
+            query_params: DictType | None = None,
+    ) -> urllib.parse.SplitResult:
         """Build URL for given information."""
         if not bucket_name and object_name:
             raise ValueError(
@@ -649,8 +724,8 @@ class BaseURL:
         )
 
         if self._aws_info:
-            url = self._build_aws_url(
-                url, bucket_name, enforce_path_style, region)
+            url = BaseURL._build_aws_url(
+                self._aws_info, url, bucket_name, enforce_path_style, region)
 
         netloc = url.netloc
         path = "/"
@@ -669,8 +744,14 @@ class ObjectWriteResult:
     """Result class of any APIs doing object creation."""
 
     def __init__(
-            self, bucket_name, object_name, version_id, etag, http_headers,
-            last_modified=None, location=None,
+            self,
+            bucket_name: str,
+            object_name: str,
+            version_id: str | None,
+            etag: str | None,
+            http_headers: HTTPHeaderDict,
+            last_modified: datetime | None = None,
+            location: str | None = None,
     ):
         self._bucket_name = bucket_name
         self._object_name = object_name
@@ -681,37 +762,37 @@ class ObjectWriteResult:
         self._location = location
 
     @property
-    def bucket_name(self):
+    def bucket_name(self) -> str:
         """Get bucket name."""
         return self._bucket_name
 
     @property
-    def object_name(self):
+    def object_name(self) -> str:
         """Get object name."""
         return self._object_name
 
     @property
-    def version_id(self):
+    def version_id(self) -> str | None:
         """Get version ID."""
         return self._version_id
 
     @property
-    def etag(self):
+    def etag(self) -> str | None:
         """Get etag."""
         return self._etag
 
     @property
-    def http_headers(self):
+    def http_headers(self) -> HTTPHeaderDict:
         """Get HTTP headers."""
         return self._http_headers
 
     @property
-    def last_modified(self):
+    def last_modified(self) -> datetime | None:
         """Get last-modified time."""
         return self._last_modified
 
     @property
-    def location(self):
+    def location(self) -> str | None:
         """Get location."""
         return self._location
 
@@ -719,7 +800,12 @@ class ObjectWriteResult:
 class Worker(Thread):
     """ Thread executing tasks from a given tasks queue """
 
-    def __init__(self, tasks_queue, results_queue, exceptions_queue):
+    def __init__(
+            self,
+            tasks_queue: Queue,
+            results_queue: Queue,
+            exceptions_queue: Queue,
+    ):
         Thread.__init__(self, daemon=True)
         self._tasks_queue = tasks_queue
         self._results_queue = results_queue
@@ -751,8 +837,13 @@ class Worker(Thread):
 
 class ThreadPool:
     """ Pool of threads consuming tasks from a queue """
+    _results_queue: Queue
+    _exceptions_queue: Queue
+    _tasks_queue: Queue
+    _sem: BoundedSemaphore
+    _num_threads: int
 
-    def __init__(self, num_threads):
+    def __init__(self, num_threads: int):
         self._results_queue = Queue()
         self._exceptions_queue = Queue()
         self._tasks_queue = Queue()
@@ -777,7 +868,7 @@ class ThreadPool:
                 self._tasks_queue, self._results_queue, self._exceptions_queue,
             )
 
-    def result(self):
+    def result(self) -> Queue:
         """ Stop threads and return the result of all called tasks """
         # Send None to all threads to cleanly stop them
         for _ in range(self._num_threads):
